# coding: utf-8
from __future__ import print_function
import argparse
import sys
import os
import platform
import shutil
from .settings import warning_dict, settings
from .preparation import welcome, create_workdir, output_log, def_res_shells
from .preparation import which, res_high_from_xyzin, res_from_mtz, res_opt
from .preparation import calculate_merging_stats, run_pdbtools
from .preparation import res_from_hklin_unmerged, check_refinement_software
from .commons import twodec, twodecname, warning_my, try_symlink
from .refinement import collect_stat_OVERALL
from .refinement import collect_stat_OVERALL_AVG
from .refinement import collect_stat_BINNED
from .graphs import matplotlib_bar, matplotlib_line, write_log_html


RES_LOW = 50


class MyArgumentParser(argparse.ArgumentParser):
    """ Helper class for `argparse`

    It adds an attribute `add_argument_with_check` that check if file
    given in argument exist but not open it.
    Inspired by `https://codereview.stackexchange.com/questions/28608/
    checking-if-cli-arguments-are-valid-files-directories-in-python`
    """
    def __is_valid_file(self, parser, arg):
        """Checks if file
        given in argument `arg` exists but does not open it.
        If not, abort.

        Args:
            self
            parser: parser of `argparse`
            arg (str): argument of `argparse`

        Returns:
            str: Name of the checked file
        """
        if not os.path.isfile(arg):
            parser.error('The file {} does not exist!'.format(arg))
        else:
            # File exists so return the filename
            return arg

    def add_argument_with_check(self, *args, **kwargs):
        """New attribute for `argparse` that checks if file
        given in argument exist but does not open it.
        """
        # Look for your FILE settings
        # type = lambda x: self.__is_valid_file(self, x) # PEP8 E731
        def type(x):
            return self.__is_valid_file(self, x)
        kwargs['type'] = type
        self.add_argument(*args, **kwargs)

    def error(self, message):
        sys.stderr.write('error: %s\n' % message)
        print('Run `ccp4-python -m pairef -h` to show the help message.')
        sys.exit(2)


# https://stackoverflow.com/questions/14117415/
# in-python-using-argparse-allow-only-positive-integers
def check_positive_int(value):
    ivalue = int(value)
    if ivalue <= 0:
        raise argparse.ArgumentTypeError("%s is an invalid positive "
                                         "int value" % value)
    return ivalue


def check_non_negative_int(value):
    ivalue = int(value)
    if ivalue < 0:
        raise argparse.ArgumentTypeError("%s is an invalid non-negative "
                                         "int value" % value)
    return ivalue


def check_positive_float(value):
    ivalue = float(value)
    if ivalue <= 0:
        raise argparse.ArgumentTypeError("%s is an invalid positive "
                                         "float value" % value)
    return ivalue


def process_arguments(input_args):
    '''Processes input arguments using `argparse`.

    Args:
        input_args (list): Input arguments

    Returns:
        list: Processed arguments
    '''
    from . import __version__
    # Input processing
    # parser = argparse.ArgumentParser(
    parser = MyArgumentParser(
        description="Automatic PAIRed REFinement protocol",
        epilog='Dependencies: CCP4 Software Suite or PHENIX containing CCTBX '
        'with Python 2.7',
        prog='ccp4-python -m pairef',
        add_help=False)

    # Arguments
    parser._optionals.title = 'optional arguments specifying input files'
    # This is a bit dirty hack but  the input files has to be added to `parser`
    # (not to a group) in order to add_argument_with_check() could work
    parser.add_argument(
        '--GUI', '--gui', dest='gui', action='store_true',
        help='Start graphical user interface (usually requires '
        "to be executed as ccp4-python, not as cctbx.python)")
    ###########################   GUI   ######################################
    if "--GUI" in input_args or "--gui" in input_args:   
        args, args_unknown = parser.parse_known_args(input_args)
        return args
    ###########################   GUI   ######################################
        
    parser.add_argument_with_check(
        '--XYZIN', '--xyzin', dest='xyzin',
        help='PDB or mmCIF file with current structure model', required=True)
    parser.add_argument_with_check(
        '--HKLIN', '--hklin', dest='hklin',
        help='MTZ file with processed diffraction data', required=True)
    parser.add_argument_with_check(
        '-u', "--unmerged", dest='hklin_unmerged',
        help='unmerged processed diffraction data file (e.g. XDS_ASCII.HKL '
        'or data_unmerged.mtz)')
    parser.add_argument_with_check(
        '--LIBIN', '--libin', dest='libin',
        help='CIF file geometric restraints')
    parser.add_argument_with_check(
        '--TLSIN', '--tlsin', dest='tlsin',
        help='input TLS file (only for REFMAC5)')
    parser.add_argument_with_check(
        '-c', "--comfile", dest='comin',  # TODO - better...
        help='configuration Com file with keywords for REFMAC5')
    parser.add_argument_with_check(
        '-d', "--def", dest='defin',
        help='configuration def file with keywords for phenix.refine')

    group1 = parser.add_mutually_exclusive_group()
    #    'optional arguments specifying refinement software')
    group1.add_argument(
        '-R', '--refmac', dest='refmac',
        help='Use REFMAC5 (default)', action='store_true')
    group1.add_argument(
        '-P', '--phenix', dest='phenix',
        help='Use phenix.refine', action='store_true')

    group2 = parser.add_argument_group('other optional arguments')
    group2.add_argument(
        '-p', "--project", dest='project', help='project name')
    group2.add_argument(
        '-r', dest='res_shells',
        help='explicit definition of high resolution shells - '
        'values must be divided using commas without any '
        'spaces and written in decreasing order, e.g. '
        '2.1,2.0,1.9')
    group2.add_argument(
        '-n', dest='n_shells',
        help='number of high resolution shells to be added step by step. '
        'Using this argument, setting of argument -s is required.',
        type=check_positive_int)
    group2.add_argument(
        '-s', "--step", dest='step',
        help='width of the added high resolution shells (in angstrom). '
        'Using this argument, setting of argument -n is required.',
        type=float)
    group2.add_argument(
        '-i', dest='res_init',
        help='initial high-resolution diffraction limit (in angstrom) '
        '- if it is not necessary, '
        'do not use this option, the script should find resolution '
        'automatically in PDB or mmCIF file',
        type=float)
    group2.add_argument(
        '-f', "--flag", dest='flag',
        help="definition which FreeRflag set will be excluded during "
        "refinement (set 0 default)", type=check_non_negative_int)
    group2.add_argument(
        '-w', "--weight", dest='weight',
        help="manual definition of weighting term (only for REFMAC5)",
        type=float)
    group2.add_argument(
        "--ncyc", dest='ncyc',
        help="number of refinement cycles that will be performed in every "
        "resolution step", type=check_positive_int)
    group2.add_argument(
        '--constant-grid', dest='constant_grid',
        help="keep the same FFT grid through the whole paired refinement "
        "(only for REFMAC5)", action='store_true')
    group2.add_argument(
        '--complete', dest='complete_cross_validation',
        help="perform complete cross-validation (use all available free "
        "reflection sets)", action='store_true')
    group2.add_argument(
        "--TLS-ncyc", "--tls-ncyc", dest='tls_ncyc',
        help="number of cycles of TLS refinement (10 cycles by default, "
        "only for REFMAC5)",
        type=check_positive_int)
    group2.add_argument(
        '--TLSIN-keep', "--tlsin-keep", dest='tlsin_keep',
        help="keep using the same TLS input file in all the refinement runs "
        "(only for REFMAC5)", action='store_true')
    group2.add_argument(
        "--open-browser", action="store_true", dest='open_browser',
        help="open web browser to show results "
        "(requires to be executed as ccp4-python, not as cctbx.python)")
    group2.add_argument(
        "-h", "--help", action="help", help="show this help message and exit")
    group2.add_argument(
        '-v', '--version', action='version', version=__version__,
         help="show version and exit")

    group3 = parser.add_argument_group(
        'optional arguments specifying structure model modification')
    group3.add_argument(
        "--prerefinement-ncyc", dest='prerefinement_ncyc',
        help="number of refinement cycles to be performed as pre-refinement "
        "of the input structure model before paired "
        "refinement (the initial high resolution limit is used). Pre-refine"
        "ment is performed by default in case of the complete cross-"
        "validation protocol. Other related options are "
        "--prerefinement-reset-bfactor, --prerefinement-add-to-bfactor, "
        "--prerefinement-set-bfactor, --prerefinement-shake-sites, and "
        "--prerefinement-no-modification. "
        "These options can be useful when the structure has "
        "been refined in another version of REFMAC5 or phenix.refine "
        "than it is currently used or when you want to reset the impact of "
        "used free reflections.",
        type=check_positive_int)
    group3.add_argument(
        "--prerefinement-reset-bfactor", dest='reset_bfactor',
        help="reset atomic B-factors of the input structure model to "
        "the mean value. This is done by default in the case of the complete"
        "cross-validation protocol.",
        # "This argument requires the argument --prerefinement_ncyc.",
        action='store_true')
    group3.add_argument(
        '--prerefinement-add-to-bfactor', dest='add_to_bfactor',
        help="add the given value to B-factors"
        " of the input structure model", type=float)
    group3.add_argument(
        "--prerefinement-set-bfactor", dest='set_bfactor',
        help="set atomic B-factors of the input structure model "
        "to the given value.",
        # "This argument requires the argument --prerefinement_ncyc.",
        type=check_positive_float)
    group3.add_argument(
        "--prerefinement-shake-sites", dest='shake_sites',
        help="randomize coordinates of the input structure model with the "
        "given mean error value. This is done by default in the case of the "
        "complete cross-validation protocol - mean error 0.25.",
        # "This argument requires the argument --prerefinement_ncyc.",
        type=check_positive_float,
        nargs='?', const=0.25)
    # nargs='?' means 0-or-1 arguments
    # const=0.25 sets the default when there are 0 arguments
    group3.add_argument(
        '--prerefinement-no-modification', dest='no_modification',
        help="do not modify the input structure model before the complete "
        "cross-validation protocol", action='store_true')

    parser.add_argument(
        '-t', dest='test', help=argparse.SUPPRESS, action='store_true')
    parser.add_argument(
        '-q', dest='quick', help=argparse.SUPPRESS, action='store_true')

    # If the arguments are not set correctly, show help and exit
    if (input_args == [__file__]) or \
       (input_args == [os.path.basename(__file__)]) or \
       (len(input_args) == 0):
        parser.print_help()  # help on empty input
        sys.exit(1)
    # Passing list of params to parse_args without script name
    if (__file__ in input_args) or \
       (os.path.basename(__file__) in input_args) or \
       (input_args[0].split("/")[-1] == "__main__.py") or \
       (input_args[0].split("/")[-1] == "pairef"):
        input_args = input_args[1:]
    args, args_unknown = parser.parse_known_args(input_args)
    # CCP4Console on Windows adds an unknown argument (abs path)\__main__.py
    for arg_unknown in args_unknown:
        if "__main__.py" not in arg_unknown:
            print("WARNING: Unknown argument: " + arg_unknown)
    # Management of conflicts
    #
    # if ((args.add_to_bfactor and
    #         not (args.complete_cross_validation or
    #              args.reset_bfactor))):
    #     parser.error("The argument --prerefinement-add-to-bfactor requires the"
    #                  " argument --complete or --prerefinement-reset-bfactor.")
    if args.complete_cross_validation and isinstance(args.flag, (int, long)):
        parser.error("It is a non-sense to use the option -f with the option "
                     "--complete.")
    if (args.no_modification and
            (args.reset_bfactor or args.add_to_bfactor or
             args.set_bfactor or args.shake_sites)):
        parser.error("It is a non-sense to use the option "
                     "--prerefinement-no-modification "
                     "with the options --prerefinement-reset-bfactor, "
                     "--prerefinement-add-to-bfactor, "
                     "--prerefinement-set-bfactor, or "
                     "--prerefinement-shake-sites.")
    if ((not args.prerefinement_ncyc and
            not args.complete_cross_validation) and
            (args.reset_bfactor or args.add_to_bfactor or
             args.set_bfactor or args.shake_sites)):
        parser.error("It is not supported to modify the input structure model "
                     "(arguments --prerefinement-reset-bfactor, "
                     "--prerefinement-add-to-bfactor, "
                     "--prerefinement-set-bfactor, or "
                     "--prerefinement-shake-sites) and not perform "
                     "pre-refinement at the initial resolution - use the "
                     "required argument --prerefinement-ncyc")
    if (args.no_modification and not args.complete_cross_validation):
        print("NOTE: The option --prerefinement-no-modification is "
              "redundant when the option --complete is not used.")
    if (args.set_bfactor and (args.reset_bfactor or args.add_to_bfactor)):
        parser.error("The option --prerefinement-set-bfactor cannot be "
                     "combined with the options --prerefinement-reset-bfactor "
                     "and --prerefinement-add-to-bfactor.")
    if args.complete_cross_validation and args.reset_bfactor:
        print("NOTE: The argument --prerefinement-reset-bfactor is "
              "redundant when the argument --complete is used - the B-factors "
              "are reseted to their average value by default.")
    if args.complete_cross_validation and args.no_modification:
        warning_my("no_modification", "Modification of the input structure "
                   "model is turned off - be sure that the input structure "
                   "model has been pertubed. If not, the results will be "
                   "biased.")
    # if ((args.reset_bfactor and
    #      not args.prerefinement_ncyc)):
    #     parser.error("The argument --prerefinement-reset-bfactor requires "
    #                  "the argument --prerefinement-ncyc.")
    if (args.step and not args.n_shells) or \
            (not args.step and args.n_shells):  # TODO
        parser.error("Both - the number and the width - of high resolution "
                     "shells (options -n and -s) "
                     "must be set in one time (not only one of "
                     "these options).")
        args.reset_bfactor = False
    # Input structure model modification - default behaviour
    if args.complete_cross_validation and not args.no_modification:
        if not (args.reset_bfactor or args.add_to_bfactor or
                args.set_bfactor or args.shake_sites):
            args.reset_bfactor = True
            args.shake_sites = 0.25
    # REFMAC5   X   phenix.refine
    if args.phenix and args.constant_grid:
        parser.error("Option --constant-grid is currently supported only for"
                     "REFMAC5, not for phenix.refine.")
    if args.phenix and args.weight:
        parser.error("Weight can be directly specifyied only if "
                     "phenix.refine is set as refinement software. Specify "
                     "keywords in a def file and use an option --def to set "
                     "parameters of phenix.refine.")
    if args.phenix and args.comin:
        parser.error("It is not possible to provide Com file for REFMAC5 when "
                     "phenix.refine is set as refinement software. Use an "
                     "option --def instead to specify keywords for "
                     "phenix.refine  or set refinement in REFMAC5 using an "
                     "option --refmac.")
    if not args.phenix and args.defin:
        parser.error("It is not possible to provide def file for phenix."
                     "refine when REFMAC5 is set as refinement software. Use "
                     "an option --comin instead to specify keywords for "
                     "REFMAC5 or set refinement in phenix.refine using an "
                     "option --phenix.")
    if args.phenix and (args.tlsin or args.tls_ncyc or args.tlsin_keep):
        parser.error("Specific options for TLS refinement are valid only for "
                     "REFMAC5. For phenix.refine, specify a refinement "
                     "strategy and TLS groups (keywords "
                     "refinement.refine.strategy and refinement.refine.adp) "
                     "in a configuration file (option --def).")
    # TLS
    if (args.tlsin_keep or args.tls_ncyc) and not args.tlsin:
        parser.error("Input TLS file must be specified (option --TLSIN) while "
                     "using the options --TLSIN-keep or --TLS-ncyc.")
    return(args)


def main(args):
    """The main function of the `pairef` module.

    Args:
        args: Input arguments processed by `argparse`
    """
    # Check software versions (matplotlib should be checked later)
    # if int(platform.python_version_tuple()[0]) != 2:
    #     sys.stderr.write("ERROR: This version of pairef module requires "
    #                      "Python 2.7 from CCTBX.\n")
    #     sys.exit(1)
    from . import __version__
    try:
        import cctbx.miller
    except ImportError:
        sys.stderr.write("ERROR: This version of pairef module requires "
                         "Python 2.7 from CCTBX.\n"
                         "It has to be executed using command:"
                         "ccp4-python -m pairef ARGUMENTS\n"
                         "or\n"
                         "cctbx.python -m pairef ARGUMENTS\n"
                         "Aborting.\n")
        sys.exit(1)

    # Decide which refinement software will be used
    if args.phenix:
        refinement = "phenix"
        refinement_name = "phenix.refine"
        from .refinement import refinement_phenix
        from .refinement import collect_stat_binned_phenix_low
    else:
        refinement = "refmac"  # REFMAC5 as default
        refinement_name = "REFMAC5"
        from .refinement import refinement_refmac
        from .refinement import collect_stat_binned_refmac_low
    settings["sh"] = False
    if args.phenix and platform.system() == 'Windows':
        settings["sh"] = True

    # Check of the needed executables - works only on Linux or Windows
    if platform.system() == 'Linux' or platform.system() == 'Windows':
        if refinement == "refmac":  # CCP4 & REFMAC5
            cryst_package = "CCP4 Software Suite"
            required_executables = ["refmac5", "baverage", "mtzdump", "sfcheck"]
            ## if args.update_waters:
            ##     required_executables.append("findwaters")
        elif refinement == "phenix":
            cryst_package = "PHENIX software suite"
            if platform.system() == 'Linux':
                required_executables = ["phenix.refine"]
            else:
                required_executables = []  # just skip the check
            # modules: from xia2.Wrappers.CCP4.Mtzdump import Mtzdump
            #          iotbx
        for required_executable in required_executables:
            if not which(required_executable) and not args.test:
                sys.stderr.write("ERROR: PAIREF requires installed `"
                                 "" + required_executable + "` (a part "
                                 "of the " + cryst_package + ") but it is not "
                                 "executable."
                                 "\nAborting.\n")
                sys.exit(1)

    # If a project name is not set, assign something
    if not args.project:
        args.project = "project"

    # Decide suffix for structure model (PDB X mmCIF) based on args.xyzin
    xyzin_suffix = args.xyzin.split(".")[-1]
    if xyzin_suffix == "mmcif" or xyzin_suffix == "cif":
        if refinement == "phenix":
            settings["pdbORmmcif"] = ".cif"
        else:  # refmac
            settings["pdbORmmcif"] = ".mmcif"
    else:
        settings["pdbORmmcif"] = ".pdb"

    # Create new working directory (name related to the project)
    workdir = create_workdir(args.project)

    # Set to write STDOUT to screen and file
    writer = output_log(sys.stdout, workdir + '/PAIREF_out.log')
    sys.stdout = writer

    versions_dict = {"refmac_version": "N/A",  # It will be found later
                     "phenix_version": "N/A",  # It will be found later
                     "pairef_version": __version__}

    # Show information about the module and input parameters
    welcome(args, versions_dict["pairef_version"])

    # Find resolution range of merged data
    res_low, res_high_mtz = res_from_mtz(args.hklin)  # uses CCTBX
    if not res_high_mtz:
        sys.stderr.write("ERROR: High-resolution limit of data "
                         "" + args.hklin + " could not be found."
                         "\nAborting.\n")
        sys.exit(1)
    if not res_low:
        warning_my("low_res", "Low resolution limit could not be found."
                   "Setting it to a value " + str(RES_LOW) + " A.")
        res_low = RES_LOW
    elif res_low > 999:
        warning_my("low_res", "Founded low resolution limit is too large "
                   "(" + str(res_low) + " A). Setting it to a value "
                   "" + str(RES_LOW) + "A.")
        res_low = RES_LOW
    else:
        print("Resolution of the merged diffraction data "
              "" + args.hklin + ": " + twodec(res_low) + "-"
              "" + twodec(res_high_mtz) + " A")

    if args.hklin_unmerged:
        # Find resolution range of unmerged data
        res_low_from_hklin_unmerged, res_high_from_hklin_unmerged = \
            res_from_hklin_unmerged(args.hklin_unmerged)
        if res_high_from_hklin_unmerged != 0 and \
                res_low_from_hklin_unmerged != float("inf"):
            print("Resolution of the unmerged diffraction data "
                  "" + args.hklin_unmerged + ": "
                  "" + twodec(res_low_from_hklin_unmerged) + "-"
                  "" + twodec(res_high_from_hklin_unmerged) + " A")

    # Set initial high resolution limit
    if args.res_init:
        print("Manual setting of initial high resolution limit will be "
              "used: " + twodec(args.res_init) + " A.")
    else:
        args.res_init = res_high_from_xyzin(
            args.xyzin, format=settings["pdbORmmcif"])
        if args.res_init < 0:
            sys.stderr.write(
                "ERROR: An attempt to determine a resolution of data which "
                "were used for refinement of the structure model "
                "" + args.xyzin + " was not successful. Please specify the "
                "initial high resolution limit manually using -i option."
                "\nAborting.\n")
            sys.exit(1)
        print("Initial high resolution limit found in the structure model "
              "" + args.xyzin + ": " + twodec(args.res_init) + " A.")

    # Check that resolution shell setting has sence
    # and determine resolution shells
    shells, n_bins_low, n_flag_sets, default_shells_definition = \
        def_res_shells(args, refinement, res_high_mtz, res_low)
    print("High resolution diffraction limits:", end=" ")
    for shell in shells[1:-1]:  # Skip the initial high resolution limit
        print(twodec(shell) + " A", end=", ")
    print(twodec(shells[-1]) + " A")  # Formatting issue

    # Set FreeRflag sets
    if args.complete_cross_validation:
        if n_flag_sets <= 2:
            sys.stderr.write(
                "Given input MTZ file " + args.hklin + " has too low number "
                "of free reflection sets (" + str(n_flag_sets) + "). k-fold "
                "cross-validation cannot be performed.\n"
                "Aborting.\n")
            sys.exit(1)
        flag_sets = range(n_flag_sets)
        if args.quick:  # Faster testing
            flag_sets = range(3)
    else:
        # Python 2.x requires the bracket (int, long)
        # https://stackoverflow.com/questions/3501382/
        # checking-whether-a-variable-is-an-integer-or-not
        if not isinstance(args.flag, (int, long)):
            args.flag = 0
        flag_sets = [args.flag]
        print(" * Data with FreeRflag set " + str(args.flag) + " will be "
              "excluded during refinement.")

    # Check the input files?

    # Copy input files to the working directory
    # and take only basename of the filenames
    in_files = ["hklin", "xyzin"]
    in_files_optional = ["libin", "comin", "defin", "tlsin"]
    for f in in_files_optional:
        if vars(args)[f]:
            in_files.append(f)
    for f in in_files:
        shutil.copy2(vars(args)[f], workdir)
        vars(args)[f] = os.path.basename(vars(args)[f])
    # Symlink HKLIN_unmerged
    if args.hklin_unmerged:
        try_symlink(os.path.join(os.getcwd(), args.hklin_unmerged),
                    os.path.join(os.getcwd(), workdir,
                                 os.path.basename(args.hklin_unmerged)))
        args.hklin_unmerged = os.path.basename(args.hklin_unmerged)
    print("")
    # Change the working directory
    os.chdir(workdir)
    print("Current working directory: " + os.getcwd())

    write_log_html(shells, [], args, versions_dict, flag_sets)
    htmlfilepath = os.path.abspath("PAIREF_" + args.project + ".html")
    print("------> RESULTS AND THE CURRENT STATUS OF CALCULATIONS ARE LISTED "
          "IN A HTML LOG FILE "
          "" + htmlfilepath)
    
    if args.open_browser and "ccp4" in sys.executable:  # cctbx.python fails
        import webbrowser
        print("Opening web browser...")
        webbrowser.open(htmlfilepath)
    print("")

    # Modification of the input structure model - Define starting XYZIN
    if args.no_modification:
        xyzin_start = args.xyzin
    else:
        if args.complete_cross_validation or args.reset_bfactor:
            if not which("baverage"):  # typically PHENIX without CCP4 paths
                from .preparation import run_bmean_iotbx
                baverage = run_bmean_iotbx(args.project, args.xyzin)
            else:  # typically refmac
                from .preparation import run_baverage
                baverage = run_baverage(args.project, args.xyzin, args.res_init)
            xyzin_start = run_pdbtools(args, baverage)
        else:
            xyzin_start = run_pdbtools(args)
    shutil.copy2(args.xyzin,
                 args.project + "_" + twodecname(shells[0]) + "A" + \
                 settings["pdbORmmcif"])
    if args.test:
        sys.exit(0)

    print("\nRefinement using " + refinement_name + ":\n")
    res_cur = shells[0]
    if args.complete_cross_validation or args.prerefinement_ncyc:
        print("   * Performing pre-refinement at "
              "" + twodec(res_cur) + " A resolution...")
    else:
        print("   * Calculating initial statistics at "
              "" + twodec(res_cur) + " A resolution...")
    for flag in flag_sets:
        if refinement == "refmac":
            results = refinement_refmac(res_cur=res_cur,
                                        res_prev=args.xyzin,
                                        res_high=shells[0],
                                        args=args,
                                        n_bins_low=n_bins_low,
                                        mode="first",
                                        res_low=res_low,
                                        res_highest=shells[-1],
                                        flag=flag,
                                        xyzin_start=xyzin_start)
            # bfac_set=bfac_set)
            versions_dict["refmac_version"] = results["version"]
        elif refinement == "phenix":
            n_bins = n_bins_low
            results = refinement_phenix(res_cur=res_cur,
                                        res_prev=args.xyzin,
                                        res_high=shells[0],
                                        args=args,
                                        n_bins=n_bins_low,
                                        mode="first",
                                        res_low=res_low,
                                        res_highest=shells[-1],
                                        flag=flag,
                                        xyzin_start=xyzin_start)
            versions_dict["phenix_version"] = results["version"]
        collect_stat_OVERALL([res_cur], args, flag, refinement)
        if args.complete_cross_validation or args.prerefinement_ncyc:
            matplotlib_line(
                shells=[shells[0]],
                project=args.project,
                statistics=["Rwork_cyc", "Rfree_cyc"],
                n_bins_low=n_bins_low,
                title=r"$\mathrm{" + twodec(shells[0]) + r"\ \AA\ -" +
                "\ flag\ " + str(flag) + "}$",
                filename_suffix="R" + str(flag).zfill(2) + "_" +
                twodecname(shells[0]) + "A_stats_vs_cycle", flag=flag,
                refinement=refinement)
            write_log_html(shells, [], args,
                           versions_dict, flag_sets, shells[0])

    if not args.complete_cross_validation:
        src = args.project + "_R" + str(flag).zfill(2) + "_Rgap.csv"
        dst = args.project + "_Rgap.csv"
        try_symlink(src, dst)

    matplotlib_line(shells=[shells[0]],
                    project=args.project,
                    statistics=["Rgap"],
                    n_bins_low=n_bins_low,
                    title=r"$\it{R}_{\mathrm{free}}-"
                    r"\it{R}_{\mathrm{work}}$",
                    filename_suffix="Rgap", flag=flag)
    shells_ready_with_res_init = [shells[0]]
    write_log_html(shells, shells_ready_with_res_init, args,
                   versions_dict, flag_sets)

    # Check which software (and which version) has been used
    # for refinement of the file XYZIN
    if not (args.complete_cross_validation or \
            args.prerefinement_ncyc):
        check_refinement_software(args, versions_dict, refinement)

    if refinement == "refmac":
        logfilename = args.project + "_R" + str(flag_sets[0]).zfill(2) + "_" \
            "" + twodecname(shells[0]) + "A_comparison" \
            "_at_" + twodecname(shells[0]) + "A.log"
        mtzfilename = \
            args.project + "_R" + str(flag_sets[0]).zfill(2) + "_" \
            "" + twodecname(shells[0]) + "A.mtz"
        bins_low = collect_stat_binned_refmac_low(
            logfilename, mtzfilename, args.hklin, n_bins_low, res_low, flag)[1]
    elif refinement == "phenix":
        pdbfilename = args.project + "_R" + str(flag_sets[0]).zfill(2) + "_" \
            "" + twodecname(shells[0]) + "A_comparison" \
            "_at_" + twodecname(shells[0]) + "A_001.pdb"
        bins_low = collect_stat_binned_phenix_low(pdbfilename, n_bins_low)[0]
    bins_low = [float(bin) for bin in bins_low]
    if not args.complete_cross_validation:
        collect_stat_BINNED([res_cur], args.project, args.hklin,
                                   n_bins_low, flag, res_low, refinement)
        if which("sfcheck"):
            res_opt(shells[0], args, refinement)
            matplotlib_line(shells=[shells[0]],
                            project=args.project,
                            statistics=["res_opt"],
                            n_bins_low=n_bins_low,
                            title="Optical resolution",
                            filename_suffix="Optical_resolution", flag=flag)
        matplotlib_line(shells=[shells[0]],
                        project=args.project,
                        statistics=["Rwork"],
                        n_bins_low=n_bins_low,
                        title=r"$\it{R}_{\mathrm{work}}$",
                        filename_suffix="Rwork", flag=flag)
        matplotlib_line(shells=[shells[0]],
                        project=args.project,
                        statistics=["Rfree"],
                        n_bins_low=n_bins_low,
                        title=r"$\it{R}_{\mathrm{free}}$",
                        filename_suffix="Rfree", flag=flag)
        matplotlib_line(shells=[shells[0]],
                        project=args.project,
                        statistics=["CCwork", "CC*"],
                        n_bins_low=n_bins_low,
                        title=r"CC$_\mathrm{work}$",
                        filename_suffix="CCwork", flag=flag)
        matplotlib_line(shells=[shells[0]],
                        project=args.project,
                        statistics=["CCfree", "CC*"],
                        n_bins_low=n_bins_low,
                        title=r"CC$_\mathrm{free}$",
                        filename_suffix="CCfree", flag=flag)
        matplotlib_line(shells=[shells[0]],
                        project=args.project,
                        statistics=["n_work", "n_free"],
                        n_bins_low=n_bins_low,
                        title="Number of reflections in resol. bins",
                        filename_suffix="No_work_free_reflections",
                        flag=flag, multiscale=True)
    write_log_html(shells, shells_ready_with_res_init, args,
                   versions_dict, flag_sets)

    for i in range(len(shells) - 1):
        # TODO: check files
        res_cur = shells[i + 1]
        res_prev = shells[i]

        # Real refinement
        print("\n   * Refining using data up to "
              "" + twodec(shells[i + 1]) + " A resolution...")
        for flag in flag_sets:
            if refinement == "refmac":
                results = refinement_refmac(res_cur=res_cur,
                                            res_prev=res_prev,
                                            res_high=shells[i + 1],
                                            args=args,
                                            n_bins_low=n_bins_low,
                                            mode="refine",
                                            res_low=res_low,
                                            res_highest=shells[-1],
                                            flag=flag)
            elif refinement == "phenix":
                n_bins += 1
                results = refinement_phenix(res_cur=res_cur,
                                            res_prev=res_prev,
                                            res_high=shells[i + 1],
                                            args=args,
                                            n_bins=n_bins,
                                            mode="refine",
                                            res_low=res_low,
                                            res_highest=shells[-1],
                                            flag=flag)
            matplotlib_line(
                shells=[res_cur],
                project=args.project,
                statistics=["Rwork_cyc", "Rfree_cyc"],
                n_bins_low=n_bins_low,
                title=r"$\mathrm{" + twodec(res_cur) + r"\ \AA\ -" +
                "\ flag\ " + str(flag) + "}$",
                filename_suffix="R" + str(flag).zfill(2) + "_" +
                twodecname(res_cur) +
                "A_stats_vs_cycle", flag=flag,
                refinement=refinement)
            write_log_html(shells, shells_ready_with_res_init, args,
                           versions_dict, flag_sets, res_cur)
            print("       Calculating statistics of the refined structure "
                  "model...", end="")
            if refinement == "refmac":
                # Statistics up to prev. res. limit
                results = refinement_refmac(res_cur=res_cur,
                                            res_prev=res_prev,
                                            res_high=shells[i],
                                            args=args,
                                            n_bins_low=n_bins_low,
                                            mode="prev_pair",
                                            res_low=res_low,
                                            res_highest=shells[-1],
                                            flag=flag)
                # Statistics for `n_bins_low` shells up to init. res. limit
                results = refinement_refmac(res_cur=res_cur,
                                            res_prev=res_prev,
                                            res_high=shells[0],
                                            args=args,
                                            n_bins_low=n_bins_low,
                                            mode="comp",
                                            res_low=res_low,
                                            res_highest=shells[-1],
                                            flag=flag)
            elif refinement == "phenix":
                # Statistics up to prev. res. limit
                results = refinement_phenix(res_cur=res_cur,
                                            res_prev=res_prev,
                                            res_high=shells[i],
                                            args=args,
                                            n_bins=n_bins-1,
                                            mode="prev_pair",
                                            res_low=res_low,
                                            res_highest=shells[-1],
                                            flag=flag)
                # Statistics for `n_bins_low` shells up to init. res. limit
                results = refinement_phenix(res_cur=res_cur,
                                            res_prev=res_prev,
                                            res_high=shells[0],
                                            args=args,
                                            n_bins=n_bins_low,
                                            mode="comp",
                                            res_low=res_low,
                                            res_highest=shells[-1],
                                            flag=flag)
            collect_stat_OVERALL(shells[:i + 2], args, flag, refinement)
            if not args.complete_cross_validation:
                # Update csv files
                symlinks_src = [
                    args.project + "_R" + str(flag).zfill(2) +
                    "_R-values.csv",
                    args.project + "_R" + str(flag).zfill(2) + "_Rgap.csv"
                    ]
                symlinks_dst = [
                    args.project + "_R-values.csv",
                    args.project + "_Rgap.csv"
                    ]
                for src, dst in zip(symlinks_src, symlinks_dst):
                    try_symlink(src, dst)
                # Optical resolution
                if which("sfcheck"):
                    res_opt(res_cur, args, refinement)
                # Statistics for high resolution shells
                n_high_resolution_shells_ready = i + 1
                for j in range(n_high_resolution_shells_ready):
                    if refinement == "refmac":
                        results = refinement_refmac(res_cur=shells[i + 1],
                                                    res_prev=shells[i],
                                                    res_high=shells[j + 1],
                                                    args=args,
                                                    n_bins_low=n_bins_low,
                                                    mode="comp",
                                                    res_low=shells[j],
                                                    res_highest=shells[-1],
                                                    flag=flag)
                    elif refinement == "phenix":
                        results = refinement_phenix(res_cur=shells[i + 1],
                                                    res_prev=shells[i],
                                                    res_high=shells[j + 1],
                                                    args=args,
                                                    n_bins=1,
                                                    mode="comp",
                                                    res_low=shells[j],
                                                    res_highest=shells[-1],
                                                    flag=flag)
        shells_ready_with_res_init = shells[:i + 2]
        print("")
        if args.complete_cross_validation:
            collect_stat_OVERALL_AVG(shells_ready_with_res_init,
                                            args.project, flag_sets)
        else:
            collect_stat_BINNED(
                shells_ready_with_res_init, args.project, args.hklin,
                n_bins_low, flag, res_low, refinement)

        print("       Updating graphs...")
        matplotlib_bar(args)
        if args.complete_cross_validation:
            matplotlib_bar(args=args, flag_sets=flag_sets,
                           ready_shells=shells_ready_with_res_init)
        else:
            if which("sfcheck"):
                matplotlib_line(shells=shells_ready_with_res_init,
                                project=args.project,
                                statistics=["res_opt"],
                                n_bins_low=n_bins_low,
                                title="Optical resolution",
                                filename_suffix="Optical_resolution",
                                flag=flag)
            matplotlib_line(shells=shells_ready_with_res_init,
                            project=args.project,
                            statistics=["Rwork"],
                            n_bins_low=n_bins_low,
                            title=r"$\it{R}_{\mathrm{work}}$",
                            filename_suffix="Rwork", flag=flag)
            matplotlib_line(shells=shells_ready_with_res_init,
                            project=args.project,
                            statistics=["Rfree"],
                            n_bins_low=n_bins_low,
                            title=r"$\it{R}_{\mathrm{free}}$",
                            filename_suffix="Rfree", flag=flag)
            matplotlib_line(shells=shells_ready_with_res_init,
                            project=args.project,
                            statistics=["CCwork", "CC*"],
                            n_bins_low=n_bins_low,
                            title=r"CC$_\mathrm{work}$",
                            filename_suffix="CCwork", flag=flag)
            matplotlib_line(shells=shells_ready_with_res_init,
                            project=args.project,
                            statistics=["CCfree", "CC*"],
                            n_bins_low=n_bins_low,
                            title=r"CC$_\mathrm{free}$",
                            filename_suffix="CCfree", flag=flag)
            matplotlib_line(shells=shells_ready_with_res_init,
                            project=args.project,
                            statistics=["n_work", "n_free"],
                            n_bins_low=n_bins_low,
                            title="Number of reflections in resol. bins",
                            filename_suffix="No_work_free_reflections",
                            flag=flag, multiscale=True)
        matplotlib_line(shells=[shells[0]],  # ???
                        project=args.project,
                        statistics=["Rgap"],
                        n_bins_low=n_bins_low,
                        title=r"$\it{R}_{\mathrm{free}}-"
                        r"\it{R}_{\mathrm{work}}$",
                        filename_suffix="Rgap", flag=flag)
        write_log_html(shells, shells_ready_with_res_init, args,
                       versions_dict, flag_sets)

    # If unmerged data are in disposal, calculate CC1/2 and CC*
    # for future graphs of CCwork, CCfree
    if args.hklin_unmerged:
        write_log_html(shells, shells, args, versions_dict, flag_sets)
        calculate_merging_stats(args.hklin_unmerged, shells, args.project,
                                bins_low, res_low_from_hklin_unmerged,
                                res_high_from_hklin_unmerged)
        matplotlib_line(shells=[shells[0]], project=args.project,
                        statistics=["Rmerge", "Rmeas", "Rpim"],
                        n_bins_low=n_bins_low, title="$\it{R}$-values",
                        filename_suffix="Rmerge_Rmeas_Rpim")
        matplotlib_line(shells=[shells[0]], project=args.project,
                        statistics=["<I/sI>", "<I>"],
                        n_bins_low=n_bins_low, title="Average intensities",
                        filename_suffix="Intensities", multiscale=True)
        matplotlib_line(shells=[shells[0]], project=args.project,
                        statistics=["Completeness", "Multiplicity"],
                        n_bins_low=n_bins_low,
                        title="Completeness and multiplicity",
                        filename_suffix="Comp_Mult", multiscale=True)
        matplotlib_line(shells=[shells[0]], project=args.project,
                        statistics=["CChalf", "CC*"], n_bins_low=n_bins_low,
                        title="Correlation coefficent", filename_suffix="CC")
        matplotlib_line(shells=[shells[0]], project=args.project,
                        statistics=["n_unique", "n_obs"],
                        n_bins_low=n_bins_low,
                        title="Number of reflections in resol. bins",
                        filename_suffix="No_reflections",
                        multiscale=True)

        if not args.complete_cross_validation:
            matplotlib_line(shells=shells_ready_with_res_init,
                            project=args.project,
                            statistics=["CCwork", "CC*"],
                            n_bins_low=n_bins_low,
                            title=r"CC$_\mathrm{work}$",
                            filename_suffix="CCwork")
            matplotlib_line(shells=shells_ready_with_res_init,
                            project=args.project,
                            statistics=["CCfree", "CC*"],
                            n_bins_low=n_bins_low,
                            title=r"CC$_\mathrm{free}$",
                            filename_suffix="CCfree")
        write_log_html(shells, shells, args, versions_dict, flag_sets,
                       ready_merging_statistics=True, done=True)
    else:
        write_log_html(shells, shells, args, versions_dict, flag_sets,
                       done=True)

    if warning_dict:
        print("\nCalculation ended.")
        print("These warning messages appeared during calculation:")
        for key in warning_dict:
            print(warning_dict[key])
    else:
        print("\nCalculation ended successfully.")
    print("\nResults are listed "
          "in logfile " + os.getcwd() + "/PAIREF_" + args.project + ".html\n")
    return


def run_pairef(input_args=None):
    """**THE LAUNCHING FUNCTION** - to process input arguments
    using :func:`launcher.process_arguments` and launch
    the :func:`launcher.main` function.

    Args:
        input_args (list): List of input arguments - parameters
    """
    # Load init setting
    # init()
    # Proccess arguments
    # (Option for specifying arguments via `input_args` is due to testing)
    if not input_args:
        input_args = sys.argv
    args = process_arguments(input_args)

    if args.gui:
        # Run GUI in PyQt
        from .gui import gui
        gui()
    else:
        # Run the protocol
        main(args)
    return
 



# coding: utf-8
import sys
from .settings import warning_dict


def twodec(var):
    """Returns number with 2 decimals as a string.
    If a float is not given, it returns a string.

    Args:
        var (float)

    Returns:
        str
    """
    try:
        i = ('{:.{prec}f}'.format(var, prec=2))
    except ValueError:
        i = str(var)
    return(i)


def twodecname(var):
    """Returns number with 2 decimals but intead of decimal point is used "-"

    Args:
        var (float)

    Returns:
        str
    """
    res = twodec(var)
    out = res.replace(".", "-")
    return(out)


def fourdec(var):
    """Returns number with 4 decimals as a string.
    If a float is not given, it returns a string.

    Args:
        var (float)

    Returns:
        str
    """
    try:
        i = ('{:.{prec}f}'.format(var, prec=4))
    except ValueError:
        i = str(var)
    return(i)


def extract_from_file(filename, searched, skip_lines, n_lines,
                      nth_word=False, not_found="stop", get_first=False):
    """Returns line(s) or word relating to the search based on
    `searched` string in the file `filename`.

    If the `filename` is not found, abort (always).
    Default behavior: The last case of `searched` string match is used
    and if the `searched` string is not found in the file `filename`,
    the program is stopped with an error message.

    Args:
        filename (str): Name of the file
        searched (str): Searched string
        skip_lines (int): Number of lines to be skipped
        n_lines (int): Number of lines that should be returned
        nth_word (bool or int): `False` if lines should be returned
            or a order in a line of the word that should be picked
        not_found (str): If the `searched` string is not found, write an error
            message and exit (if `not_found="stop"`, default option)
            or (if `not_found="N/A"`) return `"N/A"` or `["N/A"]`.
        get_first (bool): Use the first case of `searched` string match.

    Returns:
        list or str:
            List containing number of lines (controlled by `n_lines`)
            with the `skip_line` offset (if `nth_word=False`) or
            picked word (`nth_word`-th word) in the `skip_lines`-th
            following line (if `nth_word=True`)
    """

    # Python 2 and 3 compatibility
    try:
        FileNotFoundError
    except NameError:
        FileNotFoundError = IOError

    try:
        # str() is needed as "TypeError: coercing to Unicode:
        #                     need string or buffer, PosixPath found"
        with open(str(filename), "r") as f:
            file_lines = f.readlines()
    except FileNotFoundError:
        sys.stderr.write("ERROR: File " + str(filename) + " was not found.\n"
                         "Aborting.\n")
        sys.exit(1)
    for i in range(len(file_lines)):
        if searched in file_lines[i]:
            j = i
            if get_first:
                break
    if "j" not in locals():
        if not_found == "stop":
            sys.stderr.write("ERROR: File " + str(filename) + " is not in a "
                             "proper format. Statistics could not be found.\n"
                             "Aborting.\n")
            sys.exit(1)
        elif not_found == "N/A":
            if not nth_word:
                lines_array = ["N/A"]
                return lines_array
            else:
                word = "N/A"
                return word
    if not nth_word:
        lines_array = file_lines[j + skip_lines:j + skip_lines + n_lines]
        return lines_array
    else:
        line = file_lines[j + skip_lines]
        try:
            word = line.split()[nth_word]
        except IndexError:
            if not_found == "stop":
                sys.stderr.write("ERROR: File " + str(filename) + " is not in "
                                 "a proper format. "
                                 "Statistics could not be found.\n"
                                 "Aborting.\n")
                sys.exit(1)
            elif not_found == "N/A":
                word = "N/A"
        return word


def warning_my(key, message):
    message = "WARNING: " + message
    if key in warning_dict:
        if not message in warning_dict[key]:
            sys.stderr.write(message + "\n")
            warning_dict[key] = warning_dict[key] + "<br />\n" + message
        # else (if message in warning_dict[key]) do nothing
    else:
        sys.stderr.write(message + "\n")
        warning_dict[key] = message
    return True


def try_symlink(src, dst):
    """Make new symlink to `src` if the `dst` file does not exist yet. If it is
    not possible to make symlinks (difficulties on Windows), just make a copy
    of the file even if exists already.

    Args:
        src (str): File name of the source
        dst (str): File name of the destination

    Returns:
        bool: True"""

    import os
    if hasattr(os, "symlink"):
        # New symlink only if it has not been made previously
        if not os.path.isfile(dst):
            os.symlink(src, dst)
    else:  # Windows
        import shutil
        shutil.copy2(src, dst)
    return True



import pytest
import os
import tempfile
import shutil
from pairef.graphs import matplotlib_bar, matplotlib_line, write_log_html
from helper import run, config, tmp_environ, AttrDict


def test_matplotlib_bar(tmp_environ):
    shutil.copy2(config("A_R-values.csv"), tempfile.gettempdir())
    args = AttrDict()
    args.project = "A"
    args.complete_cross_validation = False
    matplotlib_bar(args, values="R-values")
    assert os.path.isfile("A_R-values.png")
    os.remove("A_R-values.csv")
    os.remove("A_R-values.png")


@pytest.mark.parametrize(["statistics", "title", "filename_suffix", "pngfile"],
                         # [(["Rwork"], r"$\it{R}_{\mathrm{work}}$",
                         [(["Rwork"], "$\it{R}_{\mathrm{work}}$",
                          "Rwork", "NK_Rwork.png"),
                          # (["Rfree"], r"$\it{R}_{\mathrm{free}}$",
                          (["Rfree"], "$\it{R}_{\mathrm{free}}$",
                           "Rfree", "NK_Rfree.png")],
                         ids=["Rwork", "Rfree"])
def test_matplotlib_line_Rvalues(tmp_environ, statistics, title,
                                 filename_suffix, pngfile):
    files_in = ["NK_R00_1-80A.csv", "NK_R00_1-70A.csv", "NK_R00_1-60A.csv",
                "NK_R00_1-50A.csv"]
    for f in files_in:
        shutil.copy2(config(f), tempfile.gettempdir())
    matplotlib_line(shells=(1.8, 1.7, 1.6, 1.5), project="NK",
                    statistics=statistics, n_bins_low=10,
                    title=title, multiscale=False,
                    filename_suffix=filename_suffix)
    assert os.path.isfile(pngfile)
    files = files_in + [pngfile]
    for f in files:
        os.remove(f)


def test_matplotlib_line_Rgap(tmp_environ):
    shutil.copy2(config("NK_Rgap.csv"), tempfile.gettempdir())
    matplotlib_line(shells=(1.8), project="NK",
                    statistics="Rgap", n_bins_low=10,
                    title="$\it{R}_{\mathrm{free}}-"
                    "\it{R}_{\mathrm{work}}$", multiscale=False,
                    # title=r"$\it{R}_{\mathrm{free}}-"
                    # r"\it{R}_{\mathrm{work}}$", multiscale=False,
                    filename_suffix="Rgap")
    assert os.path.isfile("NK_Rgap.png")
    os.remove("NK_Rgap.csv")
    os.remove("NK_Rgap.png")


def test_write_log_html(tmp_environ):
    versions_dict = {"refmac_version": "N/A",
                     "pairef_version": "N/A"}
    args = AttrDict()
    args.xyzin = "fake.pdb"
    args.hklin = "fake.mtz"
    args.hklin_unmerged = "fake.HKL"
    args.libin = "fake.mtz"
    args.comin = "fake.com"
    args.tlsin = "fake.tls"
    args.defin = None
    args.phenix = None
    args.project = "A"
    args.ncyc = 11
    args.weight = None
    args.tlsin_keep = None
    args.tls_ncyc = None
    args.flag = 0
    args.add_to_bfactor = None
    args.complete_cross_validation = False
    args.no_modification = None
    args.prerefinement_ncyc = None
    args.reset_bfactor = None
    args.add_to_bfactor = None
    args.set_bfactor = None
    args.shake_sites = None
    args.constant_grid = None
    shutil.copy2(config("A_R-values.csv"), tempfile.gettempdir())
    shells_ready_with_res_init = (1.8, 1.7, 1.6)
    shells = (1.8, 1.7, 1.6, 1.5)
    flag_sets = range(20)
    htmlfilename_in_progress = write_log_html(
        shells, shells_ready_with_res_init, args, versions_dict, flag_sets)
    assert os.path.isfile(htmlfilename_in_progress)
    with open(htmlfilename_in_progress, "r") as htmlfile_in_progress:
        htmlfile_in_progress_content = htmlfile_in_progress.read()
    assert "Calculations are still in progress" in htmlfile_in_progress_content
    assert "</html>" in htmlfile_in_progress_content
    os.remove(htmlfilename_in_progress)

    htmlfilename_done = write_log_html(
        shells, shells, args, versions_dict, flag_sets, done=True)
    assert os.path.isfile(htmlfilename_done)
    with open(htmlfilename_done, "r") as htmlfile_done:
        htmlfile_done_content = htmlfile_done.read()
    assert "Calculations are still in progress" not in htmlfile_done_content
    assert "</html>" in htmlfile_done_content
    os.remove(htmlfilename_done)
    os.remove("A_R-values.csv")
    # os.remove("styles.css")



# coding: utf-8
import os
import sys
import matplotlib
matplotlib.use('agg')  # TKinter makes problems, agg should work
import matplotlib.pyplot as plt
from matplotlib.ticker import MaxNLocator
from matplotlib.lines import Line2D
import numpy as np
from collections import namedtuple
import platform
import shutil
import cgi
import warnings
from .commons import twodec, twodecname, fourdec
from .preparation import which
from .settings import warning_dict, date_time


def xticklabels_compress(list, n_max=13, depth=1):
    """ If there are more than `n_max` bins, do not show all the labels
    in list.

    Args:
        list (list): containing labels for a graph
        n_max (int): maximal allowed number of values in list that are not ""
        depth (int)

    Returns:
        list: containing labels for a graph - compressed
    """
    if len(list) > 88:
        return list  # long lists cause problems
    if len([label for label in list if label is not ""]) >= n_max:
        if len(list) % 2 == 0:
            i = int(pow(2, depth - 1) + 1)  # i = 2 in case of depth == 1
            list[1] = ""
        else:
            i = int(pow(2, depth - 1))  # i = 1 in case of depth == 1
        erase = True
        while i < len(list) - 1:
            if erase:
                list[i] = ""
                erase = False
            else:
                erase = True
            i = i + 1 * depth
        list[- depth - 1] = ""
    # If it was not enough, do it again, recursively
    if len([label for label in list if label is not ""]) >= n_max:
        list = xticklabels_compress(list=list, n_max=n_max, depth=2*depth)
    return list


def matplotlib_bar(args, values="R-values", flag_sets=[], ready_shells=[]):
    """Plots and saves a bar chart using `matplotlib`.

    If `flag_sets` is an empty
    list, it is assumed that the values are saved in the file
    `args.project_values.csv`.

    In the other case, a chart showing results
    of complete cross-validation is ploted; the needed values
    are picked from files `project_RXX_values.csv` where `XX` is a number
    of a flag.

    Args:
        args (parser): Input arguments (including e. g. name of the project) \
                       parsed by `argparse` via function process_arguments()
        values (str): expected value: `"R-values"` (not ready yet: \
                      `"CC-values"`)
        flag_sets (list): List of free reflection flag sets (int)
        ready_shells (list)

    Returns:
        str: Name of the PNG file containing the chart.
    """
    # Graph setting
    plt.rcParams["font.family"] = "serif"
    plt.rcParams["mathtext.fontset"] = "cm"
    plt.rcParams["legend.loc"] = 'best'
    plt.rcParams["legend.fontsize"] = 'large'
    fig, ax = plt.subplots()
    fig.patch.set_facecolor('white')
    # fig.tight_layout()
    bar_width = 0.35
    opacity = 1
    ax.set_ylabel(r'$\Delta \it{R}$', rotation=0, fontsize=14)

    xticklabels_list = []
    values_work_list = []
    values_free_list = []
    errors_work_list = []
    errors_free_list = []

    def pick_work_free_from_csv_line(line, values_work_list, values_free_list,
                                     errors_work_list, errors_free_list,
                                     errors=False):
        if line.lstrip()[0] == "#":  # If it is a comment, do not load data
            continue_sign = True
            return values_work_list, values_free_list, errors_work_list, errors_free_list, continue_sign
        continue_sign = False
        try:
            values_work_list.append(float(line.split()[3]))
        except ValueError:
            values_work_list.append(None)
        try:
            values_free_list.append(float(line.split()[6]))
        except ValueError:
            values_free_list.append(None)
        if errors:  # return also standard error of mean
            try:
                errors_work_list.append(float(line.split()[9]))
            except ValueError:
                errors_work_list.append(float('nan'))            
            try:
                errors_free_list.append(float(line.split()[10]))
            except ValueError:
                errors_free_list.append(float('nan'))
        else:
            errors_work_list.append(float('nan'))
            errors_free_list.append(float('nan'))
        return values_work_list, values_free_list, errors_work_list, errors_free_list, continue_sign

    if flag_sets:
        # Chart for the complete cross-validation
        # (differences of statistics depending on to various FreeRflag sets)
        pngfilename = args.project + "_" + values + "_complete_" \
            "" + twodecname(ready_shells[-1]) + "A.png"
        # Load data - statistics relating to individual free refl. sets
        for flag in flag_sets:
            xticklabel = str(flag)
            xticklabels_list.append(xticklabel)
            csvfilename = args.project + "_R" + str(flag).zfill(2) + "_" \
                "" + values + ".csv"
            with open(csvfilename, "r") as csvfile:
                last_line = csvfile.readlines()[-1]
            values_work_list, values_free_list, errors_work_list, \
                errors_free_list, continue_sign = \
                pick_work_free_from_csv_line(
                    last_line, values_work_list, values_free_list, 
                    errors_work_list, errors_free_list)
        xticklabels_list = xticklabels_compress(xticklabels_list, n_max=21)
        # Count numbers of increases and decreases
        values_work_positive = sum(1 for i in values_work_list if float(i) > 0)
        values_work_negative = sum(1 for i in values_work_list if float(i) < 0)
        values_work_zero = sum(1 for i in values_work_list if float(i) == 0)
        values_free_positive = sum(1 for i in values_free_list if float(i) > 0)
        values_free_negative = sum(1 for i in values_free_list if float(i) < 0)
        values_free_zero = sum(1 for i in values_free_list if float(i) == 0)
        # Load data - average statistics
        csvfilename = args.project + "_" + values + ".csv"
        with open(csvfilename, "r") as csvfile:
            last_line = csvfile.readlines()[-1]
        values_work_list, values_free_list, errors_work_list, \
            errors_free_list, continue_sign = pick_work_free_from_csv_line(
                last_line, values_work_list, values_free_list,
                errors_work_list, errors_free_list, errors=True)
        xticklabels_list.append("avrg")
        # Prepare lists of colors
        color1 = ["#0065BD"] * len(flag_sets) + ["#156570"]
        color2 = ["#6AADE4"] * len(flag_sets) + ["#00B2A9"]
        # Define labels and graph title
        ax.set_xlabel(r'$\mathrm{Free\ reflection\ set}$', fontsize=14)
        if values == "R-values":
            values_work_label = r'$\it{R}_{\mathrm{work}} ( ' \
                '' + str(values_work_positive) + r'\uparrow , ' \
                '' + str(values_work_negative) + r'\downarrow , ' \
                '' + str(values_work_zero) + r'\rightarrow' \
                ')$'
            values_free_label = r'$\it{R}_{\mathrm{free}} ( ' \
                '' + str(values_free_positive) + r'\uparrow , ' \
                '' + str(values_free_negative) + r'\downarrow , ' \
                '' + str(values_free_zero) + r'\rightarrow' \
                ')$'
            values_abb = "R"
        elif values == "CC-values":
            values_work_label = r'CC$_\mathrm{work}$'
            values_free_label = r'CC$_\mathrm{free}$'
            values_abb = "CC"
        graph_title = r"$\mathrm{" + twodec(ready_shells[-2]) + r"\AA}" \
            r"\rightarrow \mathrm{" + twodec(ready_shells[-1]) + r"\AA}$"

    else:
        # Define labels
        if values == "R-values":
            values_work_label = r'$\it{R}_{\mathrm{work}}$'
            values_free_label = r'$\it{R}_{\mathrm{free}}$'
            values_abb = "R"
        elif values == "CC-values":
            values_work_label = r'CC$_\mathrm{work}$'
            values_free_label = r'CC$_\mathrm{free}$'
            values_abb = "CC"
        ax.set_xlabel(r'$\mathrm{Resolution\ step\ (\AA})$', fontsize=14)
        # Chart showing differences of statistics depending on resolution
        csvfilename = args.project + "_" + values + ".csv"
        pngfilename = args.project + "_" + values + ".png"
        # Define graph title and bar colors
        if args.complete_cross_validation:
            graph_title = r'$\mathrm{Differences\ of\ overall\ } ' \
                '' + values_abb + r'$-$\mathrm{values\ averaged\ over\ free\ sets}$'
            color1 = "#156570"
            color2 = "#00B2A9"
            errors = True
        else:
            graph_title = r'$\mathrm{Differences\ of\ overall\ } ' \
                '' + values_abb + r'$-$\mathrm{values}$'
            color1 = "#0065BD"
            color2 = "#6AADE4"
            errors = False
        # Load data
        with open(csvfilename, "r") as csvfile:
            for line in csvfile.readlines():
                values_work_list, values_free_list, errors_work_list, \
                    errors_free_list, continue_sign = \
                    pick_work_free_from_csv_line(
                        line, values_work_list, values_free_list,
                        errors_work_list, errors_free_list, errors)
                if continue_sign:
                    continue
                xticklabel = line.split()[0]
                # xticklabel = xticklabel.replace("A", r"\AA")
                # angstroem units are mentioned in the x-axis label
                xticklabel = xticklabel.replace("A", "")
                xticklabel = xticklabel.replace("->", r"\rightarrow")
                xticklabel = r"$\mathrm{" + xticklabel + "}$"
                xticklabels_list.append(xticklabel)
        xticklabels_list = xticklabels_compress(xticklabels_list, n_max=21)

    n_groups = len(xticklabels_list)
    index = np.arange(n_groups)  # NumPy
    # Plot the chart
    bar1 = ax.bar(index, values_work_list, bar_width,
                  alpha=opacity, color=color1, linewidth=0,
                  label=values_work_label,
                  yerr=errors_work_list, capsize=5, ecolor='orange')
    bar2 = ax.bar(index + bar_width, values_free_list, bar_width,
                  alpha=opacity, color=color2, linewidth=0,
                  label=values_free_label,
                  yerr=errors_free_list, capsize=5, ecolor='orange')  # , hatch="/")
    ax.set_title(graph_title, fontsize=16, y=1.04)
    ax.set_xticks(index + bar_width/2)
    ax.set_xticklabels(xticklabels_list)
    ax.legend()
    if len(xticklabels_list) < 4:
        xticklabels_rotation = 0
    elif len(xticklabels_list) < 9:
        xticklabels_rotation = 30
    else:
        xticklabels_rotation = 90
    plt.setp(ax.get_xticklabels(), rotation=xticklabels_rotation,
             horizontalalignment='center', fontsize=14)
    with warnings.catch_warnings():
        # There is a bug in matplotlib 1.x.x
        # https://github.com/matplotlib/matplotlib/issues/5209
        warnings.simplefilter(action='ignore', category=FutureWarning)
        plt.savefig(pngfilename, bbox_inches="tight", dpi=96)
        # plt.savefig(pngfilename + ".eps", bbox_inches="tight", format="eps")
    plt.clf()
    plt.close('all')
    return pngfilename


def matplotlib_line(shells, project, statistics, n_bins_low, title, flag=0,
                    multiscale=False, filename_suffix="", refinement="refmac"):
    """Plots statistics values (choice by `statistics`)
    `project+"_"+twodecname(shells[*])+"A.csv"`.
    Generate and save plot `project+"_"+statistic+".png` using `matplotlib`.

    Args:
        shells (list): containing `float`
        project (str): Name of the project
        statistics (list): List of names of statistics to be plotted (`str`)
        n_bins_low (int)
        title (str)
        flag (int)
        multiscale (bool): Use 2 different y-axis for data lines
        filename_suffix (str)
        refinement (str): "refmac" or "phenix"

    Returns:
        str:
            Name of the generated file with plot
            (`project+"_"+statistic+".png`)
    """
    # Graph setting
    plt.rcParams["font.family"] = "serif"
    plt.rcParams["mathtext.fontset"] = "cm"
    # Legend setting (Rgap and multiscale graph have another setting)
    legend_loc = "center left"
    plt.rcParams["legend.fontsize"] = 'large'
    legend_bbox_to_anchor = (1.04, 0.5)
    fig, ax = plt.subplots()
    fig.patch.set_facecolor('white')
    dpi = 96
    ax.set_xlabel(r'Resolution ($\mathrm{\AA}$)')
    ax.set_ylabel('')
    ax.grid(color='#DCDCDC')
    xticklabels_rotation = 0

    # Load xtickslabels from a model refined up to the highest resolution
    xshell_list = []
    xticklabels_list = []

    # Set filename of output PNG and graph title
    if filename_suffix:
        pngfilename = project + "_" + filename_suffix + ".png"
    else:
        pngfilename = project + "_" + title + ".png"
    graph_title = title

    # Set markers and colors
    markers = Line2D.filled_markers
    markers_cycle = 0
    # Colors are inspired by https://davidmathlogic.com/colorblind/
    # Gray color was added and blue colors were modified using
    # CTU graphical manual
    # colors = ["#0065BD", "#9B9B9B", "#0B672A", "#44AA99", "#6AADE4",
    #           "#DDCC77", "#CC6677", "#AA4499", "#882255", "#000000"]
    colors = ["#808080", "#6AADE4", "#DC267F", "#FE6100", "#0065BD", "#FFB000"]
    colors_cycle = 0
    lns = []  # List that will contain lists of datalines, needed for
              # multiscale graphs

    for statistic in statistics:
        # if "Rwork" in statistics or "Rfree" in statistics
        #    or "CCwork" in statistics or "CCfree" in statistics:
        if statistic == "Rwork" or statistic == "Rfree" or \
                statistic == "CCwork" or statistic == "CCfree":
            values_list_list = []
            if statistic == "Rwork":
                column = 6
                ax.axhline(y=0.42, color='r', linestyle='-')
            if statistic == "Rfree":
                column = 7
                ax.axhline(y=0.42, color='r', linestyle='-')
            if statistic == "CCwork":
                column = 8
            if statistic == "CCfree":
                column = 9
            # graph_title = '$\it{' + statistic.replace(statistic[-4:], "") +
            # '}_{\mathrm{' + statistic[-4:] + '}}$'

            csvfilename = project + "_R" + str(flag).zfill(2) + "_" \
                "" + twodecname(shells[-1]) + "A.csv"
            with open(csvfilename, "r") as csvfile:
                for line in csvfile.readlines():
                    if line.lstrip()[0] == "#":  # If it is a comment,
                        continue                 # do not load data
                    xshell_list.append(int(line.split()[0]))
                    xticklabels_list.append(twodec(float(line.split()[3])))
            xticklabels_list = xticklabels_compress(xticklabels_list)

            # Load statistic relating to the structure models
            for i in range(len(shells)):
                csvfilename = project + "_R" + str(flag).zfill(2) + "_" \
                    "" + twodecname(shells[i]) + "A.csv"
                values_list_list.append([])
                with open(csvfilename, "r") as csvfile:
                    for line in csvfile.readlines():
                        if line.lstrip()[0] == "#":  # If it is a comment,
                            continue                 # do not load data
                        try:
                            values_list_list[i].append(
                                float(line.split()[column]))
                        except ValueError:
                            values_list_list[i].append(None)
                # Make the `markers` and `colors` lists cycled inf.
                if i + 1 - markers_cycle > len(markers):
                    markers_cycle = markers_cycle + len(markers)
                if i + 1 - colors_cycle > len(colors):
                            colors_cycle = colors_cycle + len(colors)
                # values_label = r'~$\mathrm{' + twodec(shells[i]) + r'\ \AA}$'
                values_label = twodec(shells[i]) + r' $\mathrm{\AA}$'

                # Insert `None` into missing values to obey
                # "ValueError: x and y must have same first dimension"
                missing_values = \
                    len(xticklabels_list) - len(values_list_list[i])
                for j in range(missing_values):
                    values_list_list[i].append(None)

                # Vertical line - the conservative high resolution limit
                plt.axvline(x=n_bins_low, color='#9B9B9B', linestyle='--')

                lns += ax.plot(xshell_list, values_list_list[i],
                               label=values_label,
                               marker=markers[i - markers_cycle], markersize=4,
                               markeredgewidth=0,
                               color=colors[i - colors_cycle])
            statistics.remove(statistic)

        elif statistic == "Rgap" or statistic == "res_opt":
            values_list = []
            if statistic == "Rgap":
                csvfilename = project + "_R" + str(flag).zfill(2) + "_Rgap.csv"
                values_column = 3
            else:  # statistic == "res_opt"
                csvfilename = project + "_Optical_resolution.csv"
                values_column = 1
            with open(csvfilename, "r") as csvfile:
                for line in csvfile.readlines():
                    if line.lstrip()[0] == "#":  # If it is a comment,
                        continue                 # do not load data
                    xticklabels_list.append(twodec(float(line.split()[0])))
                    values_list.append(float(line.split()[values_column]))
            xticklabels_list = xticklabels_compress(xticklabels_list)
            xshell_list = range(len(values_list))
            values_label = title
            legend_loc = "best"
            legend_bbox_to_anchor = None
            lns += ax.plot(xshell_list, values_list, label=values_label,
                           marker="o", markersize=4,
                           markeredgewidth=0, color=colors[1])
            statistics.remove(statistic)

        elif statistic == "Rwork_cyc" or statistic == "Rfree_cyc":
            values_list = []
            if statistic == "Rfree_cyc":
                values_column = 2  # for refmac
                values_range = (20, 26)  # for phenix
                values_label = r'$\it{R}_\mathrm{free}$'
                color = "#6AADE4"
            else:  # Rwork
                values_column = 1  # for refmac
                values_range = (27, 33)  # for phenix
                values_label = r'$\it{R}_\mathrm{work}$'
                color = "#0065BD"
            prefix = project + "_R" + str(flag).zfill(2) + "_" + \
                    twodecname(shells[-1]) + "A"
            if refinement == "refmac":
                logfilename = prefix + ".log"
                with open(logfilename, "r") as logfile:
                    lines = logfile.readlines()
                for i in range(len(lines)):
                    if "    Ncyc    Rfact    Rfree     FOM      -LL     " \
                            "-LLfree  rmsBOND  zBOND rmsANGL  zANGL rmsCHIRAL $$" \
                            in lines[i]:
                        j = i + 2
                while lines[j].split()[0] != '$$':
                    values_list.append(float(lines[j].split()[values_column]))
                    if not len(xticklabels_list) == len(xshell_list):
                        xticklabels_list.append(lines[j].split()[0])
                    j = j + 1
            elif refinement == "phenix":
                xticklabels_rotation = 90
                logfilename = prefix + "_001.log"
                with open(logfilename, "r") as logfile:
                    lines = logfile.readlines()
                for i in range(len(lines)):
                    if " stage r-work r-free bonds angles " \
                            "b_min b_max b_ave n_water shift" \
                            in lines[i]:
                        j = i + 1
                if "j" not in vars():
                    sys.stderr.write("ERROR: File " + str(logfilename) + " is "
                                     "not in a proper format. "
                                     "Statistics could not be found.\n"
                                     "Aborting.\n")
                    sys.exit(1)
                while lines[j].split()[-1][-1] != '-':  # until hline ---------
                    values_list.append(float(
                        lines[j][values_range[0]:values_range[1]]))
                    if not len(xticklabels_list) == len(xshell_list):
                        xticklabel = lines[j][0:11]
                        xticklabel.strip()
                        xticklabels_list.append(xticklabel)
                    j = j + 1
            xticklabels_list = xticklabels_compress(xticklabels_list, n_max=21)
            xshell_list = range(len(values_list))
            legend_loc = "best"
            legend_bbox_to_anchor = None
            lns += ax.plot(xshell_list, values_list, label=values_label,
                           marker="o", markersize=5,
                           markeredgewidth=0, color=color)
            ax.set_xlabel(r'Cycle')
            # dpi=64

    if (os.path.isfile(project + "_merging_stats.csv") and statistics) \
            or "n_work" in statistics or "n_free" in statistics:
        if "graph_title" not in locals():
            graph_title = title
        if "pngfilename" not in locals():
            pngfilename = project + "_" + title + ".png"
        if "n_work" in statistics or "n_work" in statistics:
            csvfilename = project + "_R" + str(flag).zfill(2) + "_" + \
                twodecname(shells[-1]) + "A.csv"
            xticklabels_column = 3
        elif os.path.isfile(project + "_merging_stats.csv"):
            csvfilename = project + "_merging_stats.csv"
            xticklabels_column = 2
        if not xshell_list and not xticklabels_list:
            with open(csvfilename, "r") as csvfile:
                for line in csvfile.readlines():
                    if line.lstrip()[0] == "#":  # If it is a comment,
                        continue                 # do not load data
                    xshell_list.append(int(line.split()[0]))
                    xticklabels_list.append(twodec(float(
                        line.split()[xticklabels_column])))
            xticklabels_list = xticklabels_compress(xticklabels_list)
        for i, statistic in enumerate(statistics):
            if statistic == "CC*" and os.path.isfile(csvfilename):
                values_label = r'CC$^*$'
                values_column = -1
                marker = "*"
                color = "g"
                linestyle = '--'
            else:
                marker = markers[i - markers_cycle]
                color = colors[i - colors_cycle]
                linestyle = '-'
                if statistic == "n_obs" and os.path.isfile(csvfilename):
                    values_label = 'No. observed refl.'
                    values_column = 3
                elif statistic == "n_unique" and os.path.isfile(csvfilename):
                    values_label = 'No. unique refl.'
                    values_column = 4
                elif statistic == "n_work" and os.path.isfile(csvfilename):
                    values_label = 'No. work refl.'
                    values_column = 4
                elif statistic == "n_free" and os.path.isfile(csvfilename):
                    values_label = 'No. free refl.'
                    values_column = 5
                elif statistic == "Multiplicity" \
                        and os.path.isfile(csvfilename):
                    values_label = statistic
                    values_column = 5
                elif statistic == "Completeness" \
                        and os.path.isfile(csvfilename):
                    values_label = statistic
                    values_column = 6
                elif statistic == "<I>" and os.path.isfile(csvfilename):
                    values_label = r'<$\it{I}$>'
                    values_column = 7
                elif statistic == "<I/sI>" and os.path.isfile(csvfilename):
                    values_label = r'<$I/\sigma(I)$>'
                    values_column = 8
                elif statistic == "Rmerge" and os.path.isfile(csvfilename):
                    values_label = r'$\it{R}_\mathrm{merge}$'
                    values_column = 9
                elif statistic == "Rmeas" and os.path.isfile(csvfilename):
                    values_label = r'$\it{R}_\mathrm{meas}$'
                    values_column = 10
                elif statistic == "Rpim" and os.path.isfile(csvfilename):
                    values_label = r'$\it{R}_\mathrm{pim}$'
                    values_column = 11
                elif statistic == "CChalf" and os.path.isfile(csvfilename):
                    values_label = r'CC$_\mathrm{1/2}$'
                    values_column = -2
                else:
                    break
            values_list = []
            with open(csvfilename, "r") as csvfile:
                for line in csvfile.readlines():
                    if line.lstrip()[0] == "#":  # If it is a comment
                        continue                 # do not load data
                    try:
                        values_list.append(float(line.split()[values_column]))
                    except ValueError:
                        values_list.append(None)
            values_list = values_list[:len(xshell_list)]
            if len(values_list) < len(xticklabels_list):
                values_list = values_list + \
                    [None]*(len(xticklabels_list) - len(values_list))
            if multiscale and i == 1:
                legend_bbox_to_anchor = (1.12, 0.5)
                ax2 = ax.twinx()
                ax2.tick_params('y', colors=color)
                lns += ax2.plot(xshell_list, values_list, label=values_label,
                                marker=marker, markersize=6,
                                linestyle=linestyle, markeredgewidth=0,
                                color=color)
            else:
                lns += ax.plot(xshell_list, values_list, label=values_label,
                               marker=marker, markersize=6,
                               linestyle=linestyle, markeredgewidth=0,
                               color=color)
    labs = [l.get_label() for l in lns]
    ax.legend(lns, labs, loc=legend_loc,
              bbox_to_anchor=legend_bbox_to_anchor)
    ax.set_title(graph_title, fontsize=16, y=1.03)
    plt.xticks(xshell_list)
    ax.set_xticklabels(xticklabels_list)
    plt.setp(ax.get_xticklabels(), horizontalalignment='center',
             rotation=xticklabels_rotation)
    plt.savefig(pngfilename, bbox_inches="tight", dpi=dpi)
    # plt.savefig(pngfilename + ".eps", bbox_inches="tight", format="eps")
    plt.clf()
    plt.close('all')
    return pngfilename


def write_log_html(shells, ready_shells, args, versions_dict, flag_sets,
                   res_cur=0, ready_merging_statistics=False, done=False):
    """Created html output log.

    Args:
        shells (list)
        ready_shells (list)
        args
        versions_dict (dict): Dictionary containing keys "refmac_version" and
                              "pairef_version"
        flag_sets (list)
        res_cur (float)
        ready_merging_statistics (bool)
        done (bool)

    Returns:
        str: Name of the created HTML file
    """

    def warning_orangebox(warning_keys, page):
        """If any of warning in `warning_keys` appears, show it on the page
        in yellow box.

        Args:
            warning_keys (list)
            page (str)

        Retruns:
            str
        """
        for warning_key in warning_keys:
            if warning_key in warning_dict:
                page += '\t\t<div class="orangebox">\n\t\t\t'
                page += warning_dict[warning_key]
                page += '\n\t\t</div>\n'
        return page

    import getpass
    import socket


    page = """<!DOCTYPE html>
<head>
    <title>PAIREF - results """ + args.project + """</title>
    <link rel="stylesheet" type="text/css" href="styles.css">\n"""
    if not done:
        page += """\t<meta http-equiv="refresh" content="15">\n"""
    page += """</head>
<body>
    <div id="header">
        <div id="title">PAIREF</div>
        <div id="subtitle">Automatic PAIRed REFinement protocol</div>
    </div>

    <h1>Results - """ + args.project + """</h1>\n"""

    if done:
        page += """\t<div class="greenbox">Calculations ended""" \
            """ successfully.</div>\n"""
    else:
        page += """\t<div class="yellowbox">\n"""
        page += """\t\t<span class="progress">""" \
            """Calculations are still in progress...</span> """
        if len(ready_shells) >= 2:
            page += "Displaying results for the " \
                "following high resolution diffraction limits: "
            for shell in ready_shells:
                page += twodec(shell) + " &#8491;, "
            page = page[:-2]  # erase the last comma and space
        page += """\n\t\t<span class="reload">""" \
            """<a href="javascript:window.location.reload(true)">""" \
            """REFRESH</a></span>\n"""
        page += "\t</div>\n"

    page += "\t<h2>Input parameters</h2>\n"
    page += "\t\t<table>\n"
    page += "\t\t<tr><td>Project:</td><td>" + args.project + "</td></tr>\n"
    page += "\t\t<tr><td>High resolution diffraction limits:</td><td>"
    for shell in shells:
        page += twodec(shell) + " &#8491;, "
    page = page[:-2]  # erase the last comma and space
    page += "</td></tr>\n"
    page += "\t\t<tr><td>Initial structure model:</td><td>" \
        "" + args.xyzin + "</td></tr>\n"
    page += "\t\t<tr><td>Merged diffraction data:</td><td>" \
        "" + args.hklin + "</td></tr>\n"
    if args.hklin_unmerged:
        page += "\t\t<tr><td>Unmerged diffraction data:</td>" \
            "<td>" + args.hklin_unmerged + "</td></tr>\n"
    if args.libin:
        page += "\t\t<tr><td>Crystallographic restraints:</td>" \
            "<td>" + args.libin + "</td></tr>\n"
    if args.tlsin:
        page += "\t\t<tr><td>Input TLS file:</td><td>" + args.tlsin
        if args.tlsin_keep:
            page += "(keep using the same TLS input file " \
                "in all the refinement runs)"
        page += "</td></tr>\n"
    if args.comin:
        page += "\t\t<tr><td>Keywords for REFMAC5:</td>" \
            "<td>" + args.comin + "</td></tr>\n"
    if args.defin:
        page += "\t\t<tr><td>Keywords for phenix.refine:</td>" \
            "<td>" + args.defin + "</td></tr>\n"
    if args.weight:
        page += "\t\t<tr><td>Weight matrix for REFMAC5:</td>" \
            "<td>" + fourdec(args.weight) + "</td></tr>\n"
    if args.tls_ncyc:
        page += "\t\t<tr><td>Number of cycles of TLS refinement:</td>" \
            "<td>" + str(args.tls_ncyc) + "</td></tr>\n"
    if args.ncyc:
        page += "\t\t<tr><td>Number of refinement cycles that " \
            "will be performed in every resolution step:</td>" \
            "<td>" + str(args.ncyc) + "</td></tr>\n"
    if args.prerefinement_ncyc:
        page += "\t\t<tr><td>Number of pre-refinement cycles that will be " \
            "performed before the paired refinement protocol:</td>" \
            "<td>" + str(args.prerefinement_ncyc) + "</td></tr>\n"
    if args.complete_cross_validation:
        page += "\t\t<tr><td><strong>" \
            "Complete cross-validation is performed.</strong></td>" \
            "<td>(across " + str(len(flag_sets)) + " " \
            "free reflection sets)</td></tr>\n"
    else:
        page += "\t\t<tr><td>Data with FreeRflag set " + str(args.flag) + " " \
            "will be excluded during refinement.</td>" \
            "<td></td></tr>\n"

    if (args.complete_cross_validation or args.no_modification or
            args.reset_bfactor or args.add_to_bfactor or args.set_bfactor or
            args.shake_sites):
        page += "\t\t<tr><td>Modification of the input structure model:\n"
        page += "\t\t\t<ul>\n"
    if args.reset_bfactor:
        page += "\t\t\t<li>Reset B-factors to the mean value</li>\n"
    if args.add_to_bfactor:
        page += "\t\t\t<li>Add value to B-factors: " + \
            twodec(args.add_to_bfactor) + "</li>\n"
    if args.set_bfactor:
        page += "\t\t\t<li>Set B-factors to the value: " + \
            twodec(args.set_bfactor) + "</li>\n"
    if args.shake_sites:
        page += "\t\t\t<li>Randomize coordinates with the " + \
            "given mean error value: " + twodec(args.shake_sites) + "</li>\n"
    if args.no_modification:
        page += "\t\t\t<li>No modification</li>\n"
    if (args.complete_cross_validation or args.no_modification or
            args.reset_bfactor or args.add_to_bfactor or args.set_bfactor or
            args.shake_sites):
        page += "\t\t\t</ul>\n"

    if args.constant_grid:
        page += "\t\t<tr><td>The same FFT grid will be kept through the " \
            "whole paired refinement.</td><td></td></tr>\n"
    page += "\t\t</table>\n"
    
    page += "\t\t<h2>Run details and program versions</h2>\n"
    page += "\t\t<table>\n"
    page += "\t\t<tr><td>Working directory:</td>" \
        "<td>" + os.getcwd() + "</td></tr>\n"
    page += "\t\t<tr><td>Run date and time:</td>" \
        "<td>" + date_time + "</td></tr>\n"
    page += "\t\t<tr><td>user@host:</td><td>" \
        "" + getpass.getuser() + "@" + socket.gethostname() + "</td></tr>\n"
    page += "\t\t<tr><td><i>PAIREF</i> version:</td>" \
        "<td>" + versions_dict["pairef_version"] + "</td></tr>\n"
    if args.phenix:
        page += "\t\t<tr><td><i>phenix.refine</i> version:</td>" \
            "<td>" + versions_dict["phenix_version"] + "</td></tr>\n"
    else:
        page += "\t\t<tr><td><i>REFMAC5</i> version:</td>" \
            "<td>" + versions_dict["refmac_version"] + "</td></tr>\n"
    page += "\t\t<tr><td>Python version and executable:</td>" \
        "<td>" + platform.python_version() + " " + \
        sys.executable + "</td></tr>\n"
    page += "\t\t<tr><td>matplotlib version:</td>" \
        "<td>" + matplotlib.__version__ + "</td></tr>\n"
    page += "\t\t</table>\n"

    # Show general warnings if there are some
    warning_keys = ["workdir", "low_res", "refinement_version_mismatch",
                    "refinement_not_before", "no_modification",
                    "mtzdump", "binning", "flags", "amb_labels"]
    page = warning_orangebox(warning_keys, page)
    if ready_shells:
        if len(ready_shells) >= 2:
            page += '\t<h2>Overall values</h2>\n'
            page = warning_orangebox(["overall_R"], page)
            page += '\t\t<div class="wrap">\n'
            page += '\t\t\t<div class="column">\n'
            page += '\t\t\t\t<a href="' + args.project + '_R-values.png">'
            page += '<img src="' + args.project + '_R-values.png' \
                '?shell=' + twodecname(ready_shells[-1]) + '" ' \
                'alt="R-values"></a><br />\n'
            page += '\t\t\t\tRaw data: <a href="' + args.project + '' \
                '_R-values.csv">' + args.project + '_R-values.csv' \
                '</a><br />\n'
            page += '\t\t\t\t<pre>\n'
            with open(args.project + "_R-values.csv", "r") as csvfile:
                page += csvfile.read()
            page += '</pre>'
            page += '\n\t\t\t\t<p class="note">Note: For each incremental ' \
                'step of resolution from X->Y, the <i>R</i>-values were ' \
                'calculated at resolution X.'
            if args.complete_cross_validation:
                page += ' Standard error of mean is shown in orange.'
            page += '</p>\n'
            page += '\t\t\t</div>\n'
            page += '\t\t\t<div class="column">\n'
            page += '\t\t\t\t<a href="' + args.project + '_Rgap.png">' \
                '<img src="' + args.project + '_Rgap.png' \
                '?shell=' + twodecname(ready_shells[-1]) + '" ' \
                'alt="Rgap"></a><br />\n'
            page += '\t\t\t\tRaw data: <a href="' \
                '' + args.project + '_Rgap.csv">' + args.project + '' \
                '_Rgap.csv</a>\n'
            page += '\t\t\t\t<p class="note">Note: The R-values were ' \
                'calculated at ' + twodec(shells[0]) + ' &#8491; resolution.' \
                '</p>\n'
            page += '\t\t\t</div>\n'
            page += '\t\t</div>\n'

            if args.complete_cross_validation:
                for shell in ready_shells[1:]:  # exclude the initial difr. l.
                    page += '\t\t<a href="' + args.project + '_R-values_com' \
                        'plete_' + twodecname(shell) + 'A.png">' \
                        '<img src="' + args.project + '_R-values_complete' \
                        '_' + twodecname(shell) + 'A.png' \
                        '?shell=' + twodecname(shell) + '' \
                        '" alt="R-values (complete cross-validation, ' \
                        '' + twodec(shell) + ' A)"></a>\n'

        # "Rfree", "CCfree", "Rwork", "CCwork", "No_work_free_refl." graphs
        if not args.complete_cross_validation:
            page += '\t<h2>Dependence on resolution</h2>\n'
            # Show relating warnings if there are some
            warning_keys = ["Nfree", "high_R", "low_R", "low_CC"]
            page = warning_orangebox(warning_keys, page)
            # Show "Rfree", "CCfree", "Rwork", "CCwork" graphs
            graphs = ["Rfree", "CCfree", "Rwork", "CCwork"]
            for i, graph in enumerate(graphs):
                pngfilename = args.project + "_" + graph + ".png"
                if os.path.isfile(pngfilename):
                    page += '\t\t<a href="' + args.project + '_' + graph + '' \
                        '.png">' \
                        '<img src="' + args.project + '_' + graph + '.png' \
                        '?shell=' + twodecname(ready_shells[-1]) + '' \
                        '" alt="' + graph + '"></a>\n'
                if i % 2 == 1:  # Display max. 2 graphs in row
                    page += "\t\t<br />\n"
            # Link to raw data
            page += '\t\tRaw data:\n'
            page += '\t\t<ul>\n'
            for shell in ready_shells:
                page += '\t\t\t<li><a href="' + args.project + '_' \
                    'R' + str(args.flag).zfill(2) + '_' \
                    '' + twodecname(shell) + 'A.csv">' + args.project + '_' \
                    'R' + str(args.flag).zfill(2) + '_' \
                    '' + twodecname(shell) + 'A.csv</a></li>\n'
            page += '\t\t</ul>\n'
            # Show No. work free reflections graph
            graph = "No_work_free_reflections"
            pngfilename = args.project + "_" + graph + ".png"
            if os.path.isfile(pngfilename):
                page += '\t\t<a href="' + args.project + '_' + graph + '' \
                    '.png">' \
                    '<img src="' + args.project + '_' + graph + '.png' \
                    '?shell=' + twodecname(ready_shells[-1]) + '' \
                    '" alt="' + graph + '"></a>\n'

    # Optical resolution
    pngfilename = args.project + "_Optical_resolution.png"
    csvfilename = args.project + "_Optical_resolution.csv"
    if os.path.isfile(pngfilename):
        page += "\t<h2>Optical resolution</h2>\n"
        page += '\t\t<a href="' + pngfilename + '">' \
            '<img src="' + pngfilename + \
            '?shell=' + twodecname(ready_shells[-1]) + '" ' \
            'alt="Optical resolution"></a><br />\n'
        page += '\t\tRaw data: <a href="' + csvfilename + '">' + \
            csvfilename + '</a>\n'

    csvfilename = args.project + "_merging_stats.csv"
    if args.hklin_unmerged and ready_merging_statistics \
            and os.path.isfile(csvfilename):
        graphs = ["CC", "Intensities", "Comp_Mult", "Rmerge_Rmeas_Rpim",
                  "No_reflections"]
        page += "\t<h2>Merging statistics</h2>\n"
        warning_keys = ["merging_stats", "CC*"]
        page = warning_orangebox(warning_keys, page)
        for graph in graphs:
            page += '\t\t<a href="' + args.project + '_' + graph + '' \
                '.png">' \
                '<img src="' + args.project + '_' + graph + '.png" alt="' \
                '' + graph + '"></a>\n'
        page += '\t\t<p>Raw data: <a href="' + args.project + '' \
            '_merging_stats.csv">' \
            '' + args.project + '_merging_stats.csv</a></p>\n'
        page += '\t\t<pre>\n'
        with open(args.project + "_merging_stats.csv", "r") as csvfile:
            page += cgi.escape(csvfile.read())
        page += '\t\t</pre>\n'

    # Statistics vs. cycle
    if ((args.complete_cross_validation or args.prerefinement_ncyc) and
            (ready_shells or res_cur)) or \
            (not args.complete_cross_validation and len(ready_shells) >= 2):
        if res_cur:
            ready_shells = ready_shells + [res_cur]
        page += "\t<h2>Statistics vs. cycle</h2>\n"
        for flag in flag_sets:
            page += '\t\t<div class="scroll">\n'
            page += '\t\t<div class="wrap">\n'
            for i, shell in enumerate(ready_shells):
                prefix = args.project + '_R' + str(flag).zfill(2) + \
                    '_' + twodecname(shell) + 'A'
                pngfilename = prefix + '_stats_vs_cycle.png'
                if args.phenix:
                    logfilename = prefix + '_001.log'
                    pdbfilename = prefix + '_001.pdb'
                    ciffilename = prefix + '_001.cif'
                else:  # refmac
                    logfilename = prefix + '.log'
                    pdbfilename = prefix + '.pdb'
                    ciffilename = prefix + '.mmcif'
                if os.path.isfile(pngfilename):
                    page += '\t\t\t<div class="column">\n'
                    page += '\t\t\t\t<a href="' + pngfilename + '">' \
                        '<img src="' + pngfilename + '" alt="' \
                        '' + twodec(shell) + 'A statistics vs. cycle, ' \
                        'flag ' + str(flag).zfill(2) + '"></a><br />\n'
                    page += '\t\t\t\t<a href="' + logfilename + '">' \
                        'Log file</a> from refinement at ' \
                        '' + twodec(shell) + ' &#8491;<br />\n'
                    page += '\t\t\t\t' \
                        'Structure model refined at ' + twodec(shell) + ' ' \
                        '&#8491; <a href="' + pdbfilename + '">PDB</a> ' \
                        '<a href="' + ciffilename + '">mmCIF</a>\n'
                    page += '\t\t\t</div>\n'
            page += '\t\t</div>\n'
            page += '\t\t</div>\n'
    page += """
\t<h2>References</h2>
\t\tPlease cite the used software:
\t\t<ul>
\t\t<li>Paired refinement under the control of <i>PAIREF</i>. M. Maly, K. Diederichs, J. Dohnalek, P. Kolenko (2020) <i>IUCrJ</i> <b>7</b></li>
"""
    if not args.phenix or "ccp4" in sys.executable.lower():
        page += "\t\t<li>Overview of the <i>CCP</i>4 suite and current developments. Collaborative Computational Project, Number 4 (2011) <i>Acta Cryst.</i> D<b>67</b>:235–242</li>\n"
    if args.phenix:
        page += "\t\t<li>Macromolecular structure determination using X-rays, neutrons and electrons: recent developments in <i>Phenix</i>. D. Liebschner, P.V. Afonine, M.L. Baker, G. Bunkóczi, V.B. Chen, T.I. Croll, B. Hintze, L.W. Hung, S. Jain, A.J. McCoy, N.W. Moriarty, R.D. Oeffner, B.K. Poon, M.G. Prisant, R.J. Read, J.S. Richardson, D.C. Richardson, M.D. Sammito, O.V. Sobolev, D.H. Stockwell, T.C. Terwilliger, A.G. Urzhumtsev, L.L. Videau, C.J. Williams, P.D. Adams (2019) <i>Acta Cryst.</i> D<b>75</b>:861-877</li>\n"
        page += "\t\t<li>Towards automated crystallographic structure refinement with <i>phenix.refine</i>. P.V. Afonine, R.W. Grosse-Kunstleve, N. Echols, J.J. Headd, N.W. Moriarty, M. Mustyakimov, T.C. Terwilliger, A. Urzhumtsev, P.H. Zwart, P.D. Adams (2012) <i>Acta Cryst.</i> D<b>68</b>:352-67</li>\n"
    else:
        page += "\t\t<li><i>REFMAC</i>5 for the refinement of macromolecular ""crystal structures. G.N. Murshudov, P. Skubak, A.A. Lebedev, N.S. Pannu, R.A. Steiner, R.A. Nicholls, M.D. Winn, F. Long, A.A. Vagin (2011) <i>Acta Cryst.</i> D<b>67</b>:355–367</li>\n"
    if which("sfcheck"):
        page += "\t\t<li><i>SFCHECK</i>: a unified set of procedures for evaluating the quality of macromolecular structure-factor data and their agreement with the atomic model. A.A. Vaguine, J. Richelle, S.J. Wodak (1999) <i>Acta Cryst.</i> D<b>55</b>:191–205</li>\n"
    page += """
\t\t<li>The <i>Computational Crystallography Toolbox</i>: crystallographic algorithms in a reusable software framework. R.W. Grosse-Kunstleve, N.K. Sauter, N.W. Moriarty, P.D. Adams (2002) <i>J. Appl. Crystallogr.</i> <b>35</b>:126–136</li>
\t\t</ul>
\t\t&nbsp;
\t\t<ul>
\t\t<li>Linking crystallographic model and data quality.  P.A. Karplus & K. Diederichs (2012) <i>Science</i> <b>336</b>:1030–3</li>
\t\t<li>Assessing and maximizing data quality in macromolecular crystallography. P.A. Karplus & K. Diederichs (2015) <i>Cur. Op. in Str. Biology</i> <b>34</b>:60–68</li>
\t\t<li>Better models by discarding data? P.A. Karplus & K. Diederichs (2013) <i>Acta Cryst.</i> D<b>59</b>:1215–1222</li>
\t\t</ul>


</body>
</html>"""

    htmlfilename = "PAIREF_" + args.project + ".html"
    with open(htmlfilename, "w") as htmlfile:
        htmlfile.write(page)

    # Styles
    cssfilepath = str(os.path.dirname(os.path.abspath(__file__))) + \
        "/static/styles.css"
    if os.path.isfile(cssfilepath):
        shutil.copy2(cssfilepath, ".")

    return htmlfilename



from setuptools import setup, find_packages
from pairef import __version__

with open('README.rst') as f:
    long_description = ''.join(f.readlines())


setup(
    name='pairef',
    version=__version__,
    description='Automatic PAIRed REFinement protocol',
    long_description=long_description,
    author='Martin Maly',
    author_email='martin.maly@fjfi.cvut.cz',
    keywords='macromolecular crystallography, research',
    license='GNU Lesser General Public License v3 (LGPLv3)',
    url='https://pairef.fjfi.cvut.cz/',
    packages=find_packages(),
    classifiers=[
        'Environment :: Console',
        'Environment :: Web Environment',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)',
        'Natural Language :: English',
        'Operating System :: POSIX :: Linux',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: MacOS',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Topic :: Scientific/Engineering :: Bio-Informatics',
        'Topic :: Scientific/Engineering :: Physics',
    ],
    zip_safe=False,
    entry_points={
        'console_scripts': [
            'pairef = pairef.launcher:run_pairef',
            'pairef-gui = pairef.gui:gui',
        ]
    },
    install_requires=['numpy', 'matplotlib'],
    package_data={'pairef': ['static/*.css']},
    data_files=[('bitmaps', ['pairef/static/pairef_logo_64.png'])],
    include_package_data=True,
    # setup_requires=['pytest-runner'],
    # tests_require=['pytest'],
)



import pytest
import platform
import sys
import os
import tempfile
import shutil
from pairef.commons import twodec, twodecname, fourdec, extract_from_file
from helper import run, config


@pytest.mark.parametrize(["value_in", "value_twodec",
                          "value_twodecname", "value_fourdec"],
                         [(2, "2.00", "2-00", "2.0000"),
                          (0.617, "0.62", "0-62", "0.6170")])
def test_twodec_twodecname_fourdec(value_in, value_twodec, value_twodecname,
                                   value_fourdec):
    assert twodec(value_in) == value_twodec
    assert twodecname(value_in) == value_twodecname
    assert fourdec(value_in) == value_fourdec


filename = config("lysozyme_1-40A.log")
filename_fake = config("foo")
searched_stats = "Resolution limits"
searched_Rfree = "Free R factor"
table_stats_i = """Resolution limits                    =     39.355     1.402
Number of used reflections           =      21761
Percentage observed                  =    97.1541
Percentage of free reflections       =     5.0028
Overall R factor                     =     0.2103
Free R factor                        =     0.2236
Average Fourier shell correlation    =     0.9155
AverageFree Fourier shell correlation=     0.9072
Overall weighted R factor            =     0.2013
Free weighted R factor               =     0.2116
Overall weighted R2 factor           =     0.2598
Free weighted R2 factor              =     0.2707
Average correlation coefficient      =     0.8746
Overall correlation coefficient      =     0.9484
Free correlation coefficient         =     0.9443
Cruickshanks DPI for coordinate error=     0.0750
DPI based on free R factor           =     0.0714
Overall figure of merit              =     0.8166
ML based su of positional parameters =     0.0512
ML based su of thermal parameters    =     1.3238
"""
table_stats_f1 = """Resolution limits                    =     39.355     1.402
Number of used reflections           =      21761
Percentage observed                  =    97.1541
Percentage of free reflections       =     5.0028
"""
table_stats_f2 = """Overall R factor                     =     0.2094
Free R factor                        =     0.2220
Average Fourier shell correlation    =     0.9245
AverageFree Fourier shell correlation=     0.9165
Overall weighted R factor            =     0.2007
Free weighted R factor               =     0.2115
Overall weighted R2 factor           =     0.2580
Free weighted R2 factor              =     0.2746
Average correlation coefficient      =     0.8794
Overall correlation coefficient      =     0.9487
Free correlation coefficient         =     0.9438
Cruickshanks DPI for coordinate error=     0.0746
DPI based on free R factor           =     0.0709
Overall figure of merit              =     0.8337
ML based su of positional parameters =     0.0484
ML based su of thermal parameters    =     1.2421
"""
table_stats_f = table_stats_f1 + table_stats_f2
Rfree_i = "0.2236"
Rfree_f = "0.2220"
file_not_found = """ERROR: File """ + str(filename_fake) + """ was not found.
Aborting.
"""


@pytest.mark.parametrize(["filename", "searched", "skip_lines", "n_lines",
                          "nth_word", "not_found", "get_first", "text_test",
                          "returncode"],
                         [(filename, searched_stats, 0, 20,
                           False, "stop", False, table_stats_f, 0),
                          (filename, searched_stats, 0, 20,
                           False, "stop", True, table_stats_i, 0),
                          (filename, searched_stats, 4, 16,
                           False, "stop", False, table_stats_f2, 0),
                          (filename, searched_stats, 5, 1,
                           4, "stop", True, Rfree_i, 0),
                          (filename, searched_stats, 5, 1,
                           4, "stop", False, Rfree_f, 0),
                          (filename, searched_stats, 5, 1,
                           999, "N/A", False, "N/A", 0),
                          (filename, searched_Rfree, 0, 1,
                           4, "stop", True, Rfree_i, 0),
                          (filename, searched_Rfree, 0, 1,
                           4, "stop", False, Rfree_f, 0),
                          (filename, searched_Rfree, 0, 1,
                           999, "N/A", False, "N/A", 0),
                          # (filename_fake, searched_stats, 5, 1,
                          #  4, "stop", False, file_not_found, 1),
                          (filename, "foooooo", 0, 2,
                           False, "N/A", False, "N/A", 0),  # ??? ["N/A"]
                          (filename, "foooooo", 0, 2,
                           5, "N/A", False, "N/A", 0)],
                         ids=["table_stats_final_all",
                              "table_stats_init_all",
                              "table_stats_final_skip",
                              "table_stats_Rfree_init",
                              "table_stats_Rfree_final",
                              "table_stats_Rfree_final_N/A_index_error",
                              "Rfree_init",
                              "Rfree_final",
                              "Rfree_final_N/A_index_error",
                              # "file_not_found_abort",
                              "searched_not_found_N/A_lines",
                              "searched_not_found_N/A_word"])
def test_extract_from_file(filename, searched, skip_lines, n_lines,
                           nth_word, not_found, get_first,
                           text_test, returncode):
    text = extract_from_file(filename=filename, searched=searched,
                             skip_lines=skip_lines, n_lines=n_lines,
                             nth_word=nth_word, not_found=not_found,
                             get_first=get_first)
    if returncode == 0:
        assert "".join(text) == text_test
